#!/usr/bin/python3
# -*- coding: utf-8 -*-
import pdb
import re
import sys
import os
from collections import OrderedDict
class ShellLog:
    RED = "\33[1;31m"
    GREEN = "\33[1;32m"
    YELLOW = "\33[1;33m"
    PURPLE = "\33[1;35m"
    BLACK = "\33[1;38m"
    END = "\33[0m"

    def __init__(self, flag_error=True, flag_debug=False, flag_info=True):
        self.flag_error = flag_error
        self.flag_debug = flag_debug
        self.flag_info = flag_info

    def error(self, fmt):
        if self.flag_error:
            print(f'''{ShellLog.RED}ERROR:{ShellLog.END} {ShellLog.YELLOW}{fmt}{ShellLog.END}''', file=sys.stderr)

    def debug(self, fmt):
        if self.flag_debug:
            print(f'''{ShellLog.PURPLE}DEBUG:{ShellLog.END} {ShellLog.BLACK}{fmt}{ShellLog.END}''')

    def info(self, fmt):
        if self.flag_info:
            print(f'''{ShellLog.GREEN}INFO:{ShellLog.END} {fmt}''')

class NormalDecl(ShellLog):
    COMPARE_STATE = {'EQUAL':'相等',
                     'ERR_TYPE':'不属于struct类型',
                     'ERR_NAME':'定义名不同',
                     'ERR_VALUE':'定义名不同',
                     'ERR_MEMBER':'成员不同',
                     'ERR_MODIFIER':'修饰不同'}
    def __init__(self, string='', name='', value='', member=[], modifier=[]):
        super(NormalDecl, self).__init__()
        self.string = string
        self.name = name
        self.value = value
        self.member = member
        self.modifier = modifier
        #记录最近的一次比较状态, 取值于COMPARE_STATE
        self.compare_state = ''
        self.compare_info = ''

    def get_source_string(self):
        return self.string

    def get_failed_info(self):
        return self.compare_info

    def update_compare_state(self, state, info=''):
        self.compare_state = state
        self.compare_info = info
        if state != 'EQUAL':
            self.debug(f'{self.COMPARE_STATE[state]}: {info}')

    def __eq__(self, other):
        if not isinstance(other, NormalDecl):
            self.update_compare_state('ERR_TYPE')
            return False
        #FIXME: this is so stupid, need some work to simplify it.
        if self.name != other.name:
            self.update_compare_state('ERR_NAME', f'{self.name} != {other.name}')
            return False
        if self.value != other.value:
            self.update_compare_state('ERR_VALUE', f'{self.value} != {other.value}')
            return False
        for member1, member2 in zip(self.member, other.member):
            if member1 != member2:
                self.update_compare_state('ERR_MEMBER', f'{member1} != {member2}')
                return False
        if self.modifier != other.modifier:
            self.update_compare_state('ERR_MODIFIER', f'{self.modifier} != {other.modifier}')
            return False
        self.update_compare_state('EQUAL')
        return True

    def __str__(self):
        string = ''
        for item in self.member:
            string += f'''
            member: {item}\n'''
        return f'''NormalDecl:
            string: {self.string}
            name: {self.name}
            {string}
            modifier: {self.modifier}'''

#TODO: 不需要array_dime， 不需要解析这些，直接对比字符即可
class VarDeclaration(ShellLog):
    COMPARE_STATE = {'EQUAL':'相等',
                     'ERR_VAR_DECL':'不符合变量类型',
                     'ERR_VAR_NAME':'变量名不同',
                     'ERR_VAR_TYPE':'变量类型不同',
                     'ERR_DIME':'数组维度不同'}
    def __init__(self, var_string='', var_type=[], var_name='', array_dime=0):
        super(VarDeclaration, self).__init__()
        self.var_string = var_string
        self.var_type = var_type
        self.var_name = var_name
        #数组维数， 0 代表非数组类型
        self.array_dime = array_dime
        #记录最近的一次比较状态, 取值于COMPARE_STATE
        self.compare_state = ''
        self.compare_info = ''

    def get_source_string(self):
        return self.var_string

    def get_failed_info(self):
        return self.compare_info

    def update_compare_state(self, state, info=''):
        self.compare_state = state
        self.compare_info = info
        if state != 'EQUAL':
            self.debug(f'{self.COMPARE_STATE[state]}: {info}')

    def __eq__(self, other):
        if not isinstance(other, VarDeclaration):
            self.update_compare_state('ERR_VAR_DECL')
            return False
        if self.var_name != other.var_name:
            self.update_compare_state('ERR_VAR_NAME', f'{self.var_name} != {other.var_name}')
            return False
        if self.var_type != other.var_type:
            self.update_compare_state('ERR_VAR_TYPE', f'{self.var_type} != {other.var_type}')
            return False
        if self.array_dime != other.array_dime:
            self.update_compare_state('ERR_DIME', f'{self.array_dime} != {other.array_dime}')
            return False

        self.update_compare_state('EQUAL')
        return True

    def __str__(self):
        return f'''VarDeclaration:
            var_string: {self.var_string}
            var_type: {self.var_type}
            var_name: {self.var_name}
            array_dime: {self.array_dime}'''

class FuncDeclaration(ShellLog):
    COMPARE_STATE = {'EQUAL':'相等',
                     'ERR_FUNC_DECL':'不符合函数类型',
                     'ERR_FUNC_NAME':'函数名称不同',
                     'ERR_RET':'函数返回值类型不同',
                     'ERR_PARA':'函数参数不同'}
    def __init__(self, func_string='', ret_type='', func_name='', para=[], is_func_pointer=False):
        super(FuncDeclaration, self).__init__()
        self.func_string = func_string
        self.ret_type = ret_type
        self.func_name = func_name
        self.para = para
        self.is_func_pointer = is_func_pointer
        self.compare_state = ''
        self.compare_info = ''

    def get_source_string(self):
        return self.func_string

    def get_failed_info(self):
        return self.compare_info

    def update_compare_state(self, state, info=''):
        self.compare_state = state
        self.compare_info = info
        if state != 'EQUAL':
            self.debug(f'{self.COMPARE_STATE[state]}: {info}')

    def __eq__(self, other):
        if not isinstance(other, FuncDeclaration):
            self.update_compare_state('ERR_FUNC_DECL')
            return False
        if not self.is_func_pointer:
            if self.func_name != other.func_name:
                self.update_compare_state('ERR_FUNC_NAME', f'{self.func_name} != {other.func_name}')
                return False
        if self.ret_type != other.ret_type:
            self.update_compare_state('ERR_RET', f'{self.ret_type} != {other.ret_type}')
            return False
        for index, (item1, item2) in enumerate(zip(self.para, other.para)):
            if item1 != item2:
                self.debug(f"para: {item1} != {item2}")
                self.update_compare_state('ERR_PARA', f'第{index+1}个参数: {item1} != {item2}')
                return False
        self.update_compare_state('EQUAL')
        return True

    def __str__(self):
        return f'''FuncDeclaration:
            func_string: {self.func_string}
            func_name: {self.func_name}
            ret_type: {self.ret_type}
            para: {self.para}
            is_func_pointer: {self.is_func_pointer}'''

class Matcher(ShellLog):
    def  __init__(self, code=""):
        super(Matcher, self).__init__()
        self.code = code
        self.match_state = False
        self.match_info = ""
        self.match_object_src = None
        self.match_object_std = None

    @staticmethod
    def adjustStandStr(text):
        #\uff1b 是中文；符号，替换中文；符号
        text = re.sub(r'[\uff1b]', ';', text)
        if text.count(';') == 0:
            text += ';'
        return text

    @staticmethod
    def trimCodeStr(str):
        keyword = ["extern", "inline", "static", "asmlinkage", "__pure", "__must_check", "__cold",
                   "__malloc", r"__printf\(\s*?\d,\s*?\d\)", r"__alloc_size\(\d\)", "return",
                   "__fix_address", "__assume_kmalloc_alignment", "__always_inline", r"__realloc_size\(\d\)",
                   r"__attribute__\(\(__aligned__\(16\)\)\)"]
        pattern = r'(\b|\s)(' + '|'.join(keyword) + r')(\s|\b|;)'
        #\3代表(\s|\b|;)
        new_str = re.sub(pattern, r"\1 \3", str, flags=re.MULTILINE)
        # re.sub 在某些情况下需要执行两遍才能移除全部的keyword, 后续或许需要修正
        while new_str != str:
            str = new_str
            new_str = re.sub(pattern, r"\1 \3", new_str, flags=re.MULTILINE)
        #移除中文字符
        str = re.sub(r'[\u4e00-\u9fff]', ' ', str)

        return str.strip()

    def get_source_string(self):
        if self.match_object_src:
            return self.match_object_src.get_source_string()
        return '无'

    def get_failed_info(self):
        if self.match_object_src:
            return self.match_object_src.get_failed_info()
        return self.match_info

    def readCode(self, header, flag_def=1):
        try :
            # 这个正则存在问题, 需要后续处理
            if flag_def:
                directive = r'(#\s*(define|ifdef|ifndef|endif|if|else|elif|pragma|include).*$)'
            else:
                directive = r'(#\s*(ifdef|ifndef|endif|if|else|elif|pragma|include).*$)'
            with open(header, 'r') as file:
                code = file.read()
            self.header = header
            code = re.sub(r'/\*.*?\*/', lambda m: '\n' * m.group(0).count('\n'), code, flags=re.DOTALL)
            code = re.sub(r'//.*$', '', code, re.MULTILINE)
            code = re.sub(directive, '', code, flags=re.MULTILINE)
            self.code = self.trimCodeStr(code)
            return True
        except:
            self.error("can't open %s" %(header))
            return False

    @staticmethod
    def calculate_paired_symbol(string, left='{', right='}'):
        return string.count(left) - string.count(right)

    #TODO: make these split or readCode easier to use
    @staticmethod
    def c_split(string):
        operators = ['+','-','*','/','%','<','>','=','!','&','^','|',',','(',')','[',']',';','~']
        #FIXME: too slow if run many times
        pattern = '|'.join(map(re.escape, operators))
        return re.split(f'({pattern})', string.strip())

    def analyse_macro_define(self, code):
        self.debug(f"macro {code}")
        match = re.search(r"#\s*define\s+(.*)\n", code)
        if not match or not match.group(1):
            self.debug(f"不符合define模式: {code}")
            return None
        name, value = [], []
        paired_count, index = 0, 0
        str_list = match.group(1).split()
        #str_list = c_split(match.group(1))
        for index in range(len(str_list)):
            item = str_list[index]
            if re.match(r'\s+$',item) or not item: continue
            paired_count += self.calculate_paired_symbol(item, '(', ')')
            for word in self.c_split(item):
                if word: name.append(word)
            if name and not paired_count:
                break;
        for index in range(index + 1, len(str_list)):
            item = str_list[index]
            if re.match(r'\s+$',item) or not item: continue
            for word in self.c_split(item):
                if word: value.append(word)

        result = NormalDecl(string=code, name=name, value=value)
        self.debug(result)
        return result

    def analyse_typedef(self, code):
        self.debug(f"typedef {code}")
        pattern = r'typedef\s+(.*?)(\w+);'
        match = re.search(pattern, code)
        if not (match and match.group(1) and match.group(2)):
            self.debug(f"不符合typedef模式: {code}")
        name, value = [], []
        for item in match.group(1).split():
            for word in self.c_split(item):
                name.append(word)
        value.append(match.group(2))
        result = NormalDecl(string=code, name=name, value=value)
        self.debug(result)
        return result

    def analyse_struct(self, code):
        self.debug(f"struct {code}")
        pattern = r'''struct (\s+\w+)? \s+ {            #匹配'struct name {', 其中group(1)是name
                      (.*) }                            #匹配'xxx }', group(2)是结构体的成员
                      ([^{}]*) ;                        #group(3)是变量名或者一些编译器修饰符'''
        match = re.search(pattern, code, re.S|re.X)
        if not match:
            self.debug(f"不符合strcut模式: {code}")
            return None
        name = match.group(1)
        if name: name = name.strip()
        member = match.group(2)
        if member:
            member = self.analyse_c_source(member)
        result = NormalDecl(string=code, name=name, member=member)
        self.debug(result)
        modifiers = match.group(3)
        if modifiers:
            m = []
            for word in self.c_split(modifiers):
                if word:  m.append(word)
            result.modifier = m
        self.debug(result)
        return result

    def analyse_enum(self, code):
        self.debug(f"enum {code}")
        pattern = r'''enum \s* (\w+)? \s* {           #匹配'enum name {', 其中group(1)是name
                      (.*) }\s*;                     #匹配'xxx }', group(2)是成员 '''
        match = re.search(pattern, code, re.S|re.X)
        if not match:
            self.debug(f"不符合enum模式: {code}")
            return None
        name = match.group(1)
        member = match.group(2)
        m = []
        for word in self.c_split(member):
            word = word.strip()
            if word:  m.append(word)
        result = NormalDecl(string=code, name=name, member=m)
        self.debug(result)
        return result

    def call_special_analyse(self, pattern_dict, groups, context):
        if not groups:
            return None
        for group, key in zip(groups, pattern_dict):
            if group and pattern_dict[key]:
                return pattern_dict[key](context)

    def analyse_c_source(self, source, name=''):
        self.debug(f"source: {source}")
        source = self.trimCodeStr(source)

        if not name:
            pattern_dict = OrderedDict([(r'#\s*define', self.analyse_macro_define),
                                        ('typedef', self.analyse_typedef),
                                        (r'struct\s*\w*\s*{', self.analyse_struct),
                                        (r'enum\s*\w*\s*{', self.analyse_enum),
                                       ])
        else:
            if type(name) == list:
                name = name[0]
            pattern_dict = OrderedDict([(rf'#\s*define\s+{name}[\s(]+', self.analyse_macro_define),
                                        (rf'typedef\s+{name}\s+', self.analyse_typedef),
                                        (rf'struct\s*{name}\s*{{', self.analyse_struct),
                                        (rf'enum\s*{name}\s*{{', self.analyse_enum),
                                       ])

        keyword = ['(' + item + ')' for item in pattern_dict]
        pattern = r'\s*' + '|'.join(keyword);
        context = []
        groups = None
        paired_count = 0
        result = []
        for line in source.split('\n'):
            line = line.strip()
            if name and result:
                break;
            if paired_count:
                context += f'{line}\n'
                paired_count += self.calculate_paired_symbol(line)
                continue

            match = re.search(pattern,line)
            if not match:
                #尝试解析变量声明
                if line:
                    if not name:
                        match = re.search(r'[\w*]+\s+[\w*]+\s*(\[.*?\])*\s*;', line)
                    else:
                        match = re.search(rf'[\w*]+\s+{name}\s*(\[.*?\])*\s*;', line)
                    if match:
                        result.append(self.analyse_var_decl(line))
                    else:
                        self.debug(f"没有寻找到匹配项：{line}")
                continue

            if groups:
                result.append(self.call_special_analyse(pattern_dict, groups, context))

            context = f'{line}\n'
            groups = match.groups()
            paired_count = self.calculate_paired_symbol(line)

        if groups:
            result.append(self.call_special_analyse(pattern_dict, groups, context))
        return result



    def analyse_var_decl(self, code, var_name=r"\w*"):
        #pattern = rf"([\w\s\*]*?)({var_name})\s*((\[[\w\+\-\*/\s]*\])*);"
        pattern = rf"([\w\s\*]*?) ({var_name}) \s* ((\[ .*? \])*);"
        match = re.search(pattern, code, re.X)
        if not match:
            self.debug("analyse_var_decl: cann't find {var_name} from {code}")
            return None
        var_type=[]
        for item in match.group(1).split():
            #处理指针
            if item[0] == "*":
                index = re.search(r"\**", item).end()
                var_type.append(item[:index])
                #防止加入空元素， 例如: item='**'
                if item[index:]:
                    var_type.append(item[index:])
            elif item[-1] == "*":
                tmp = re.search(r"(\w*)(\**)", item)
                var_type.append(tmp.group(1))
                var_type.append(tmp.group(2))
            else:
                var_type.append(item)
        #处理数组, FIXME:无法判断多维数组的大小，只能判断数组的维数
        array = match.group(3)
        array_dime = 0
        if array:
            #计算数组维数
            array_dime = array.count('[')
        var = VarDeclaration(var_string=match.group(0), var_type=var_type, var_name=match.group(2),
                             array_dime=array_dime)
        return var

    def analyse_func_decl(self, code, func_name=r"\w+?", func_pointer_name=r"\w*?", flag_func_pointer=0, level=0):
        para_key_words=r"(\s|\b)(int|long|char|const|struct|double|short|enum|union|float|signed)(\s|\b)"
        if flag_func_pointer:
            pattern = rf"(\w[\w\s\*]*?)\s*(\**)\(\s*?\*\s*?({func_pointer_name})\)\s*\(([^;]*)\)[^{{}}]*?;"
        else:
            pattern = rf"(\w[\w\s\*]*?)\s+(\**)({func_name})\s*\(([^;{{}}]*)\)[^{{}}]*?(;|{{)"
        match = re.search(pattern, code, flags=re.DOTALL)
        if  match == None:
            self.debug(f"error2: can't extrace siguature. code: {code}, func_name: {func_name}")
            return None
        self.debug(f"code: {code}, func_name: {func_name}")

        ret = match.group(1).strip()
        count_parentheses = 0
        child_para = ''
        result = FuncDeclaration(func_string=match.group(0), is_func_pointer=flag_func_pointer)

        if match:
            pointer = match.group(2).strip()
            if pointer:
                ret = ret + " " + pointer
            func_name = match.group(3).strip()
            paramstr = match.group(4).strip()
            params = []
            #analyse_para(paramstr)
            if paramstr and paramstr != 'void':
                for item in paramstr.split(','):
                    if count_parentheses:
                        count_parentheses += item.count('(')
                        count_parentheses -= item.count(')')
                        child_para += f",{item}"
                        if (count_parentheses == 0):
                            params.append(self.analyse_func_decl(child_para + ";", flag_func_pointer = 1, level=level+1))
                            child_para = ''
                    else:
                        func_pointer = re.search(r".*\(\s*?\*.*?\)\s*?\(", item)
                        if func_pointer:
                            count_parentheses += item.count('(')
                            count_parentheses -= item.count(')')
                            child_para += item
                        else:
                            #FIXME:此处匹配多维指针存在一些问题，考虑同analyse_var_decl功能合并
                            start = item.split()[:-1]
                            end = item.split()[-1]
                            if  end[0] == "*":
                                type = " ".join(start).strip()+" *"
                            elif re.search(para_key_words, end, flags=re.DOTALL):
                                type =  " ".join([item]).strip()
                            elif start == []:
                                type = " ".join([item]).strip()
                            else:
                                type = " ".join(start).strip()
                            params.append(type)

            result.ret_type = ret
            result.func_name = func_name
            result.para = params
            self.debug(f"level {level} code: {result.func_string}")
            self.debug(f"level {level} ret_type: {result.ret_type}, func_name: {result.func_name}, para: {result.para}, pointer: {result.is_func_pointer}")
            return result
        else:
            self.debug(f"error: failed to analyse {str}")
            return None
    def update_match_state(self, state, info=''):
        self.match_info = info
        self.match_state = state
        if not state:
            self.error(f'{self.header}:{info}')
        return state

    def match(self, funcstr):
        flag_func_pointer = False
        flag_var = False
        self.match_state = False
        #FIXME:此处会丢失一些变量的类型信息, 比如static
        funcstr = self.trimCodeStr(funcstr)
        funcstr = self.adjustStandStr(funcstr)
        #尝试常规函数
        standard = self.analyse_func_decl(funcstr)
        if standard == None:
            #尝试函数指针
            standard = self.analyse_func_decl(funcstr, flag_func_pointer=True)
            if standard == None:
                #尝试变量
                standard = self.analyse_var_decl(funcstr)
                if standard == None:
                    #错误类型2.1：标准文件的输入无法解析，不属于函数/变量的声明
                    return self.update_match_state(False, f'无法解析标准文件输入： {funcstr}')
                else:
                    flag_var = True
            else:
                flag_func_pointer = True

        self.match_object_std = standard

        if not flag_var:
            src = self.analyse_func_decl(self.code, flag_func_pointer=flag_func_pointer,\
                                         func_pointer_name=standard.func_name, func_name=standard.func_name)
        else:
            src = self.analyse_var_decl(self.code, standard.var_name)

        self.match_object_src = src

        if src == None:
            #错误类型2.2：无法从检测文件中获取(函数/变量)的(声明/定义)
           return self.update_match_state(False ,rf'''无法从检测文件中获取定义{funcstr}.''')
        if src != standard:
            #错误类型2.3：与标准定义不匹配
            self.debug(f"src != standard ")
            self.debug(f"src: {src}")
            self.debug(f"standard: {standard}")
            return self.update_match_state(False ,f'与标准定义不匹配.定义: {funcstr}, header: {self.header} {src.compare_info}')
        return self.update_match_state(True)
